/*
 * 	usq.c - CP/M compatible file unsqueezer utility
 * this one for TRSDOS6/LDOS
 * compile w/AZTEC C
 * define TRS4 for mod 4 version
 */

#include <stdio/h>
#include <ctype/h>

#define TRUE 1
#define FALSE 0
#define ERROR (-1)
#define PATHLEN	40	/* Number of characters allowed in pathname */
#define OK 0
#define LF 10
#define CR 13
#define RECOGNIZE 0xFF76	/* unlikely pattern */
#define DLE 0x90		/* repeat byte flag */
#define SPEOF 256		/* special endfile token */
#define NUMVALS 257		/* 256 data values plus SPEOF*/
#define LARGE 30000

#ifdef VAX   /* then we don't want 32 bit integers */

typedef short INT;
typedef unsigned short UNSIGNED;

#else   /*  16 bit machines  */

typedef int INT;
typedef unsigned UNSIGNED;

#endif

struct _sqleaf {		/* Decoding tree */
	INT _children[2];	/* left, right */
};
struct _sqleaf Dnode[NUMVALS - 1];


INT Bpos;		/* last bit position read */
INT Curin;		/* last byte value read */
INT Repct;		/* Number of times to return value */
INT Value;		/* current byte value or EOF */

INT MakeLCPathname=TRUE;	/* translate pathname to lc if all caps */
INT Nlmode=FALSE;		/* zap lf's if true  chgd for TRSDOS */
INT Inbackground = FALSE;

INT getcr(), getuhuff(), portgetw();
char outdrv[3]; /* save output drive spec */
unsigned dispcnt,linecnt;   /* how much of file to view */
char inparg[PATHLEN];

main(argc, argv)
char *argv[];
{
	register char *cp;
	register INT npats=0;
	char **patts;
	INT n, errorstat,i,c;
	*outdrv = '\0'; /* default to write on any drive */
	dispcnt =0;

	/* comment out unix stuff.....
		if (signal(SIGINT, SIG_IGN)==SIG_IGN)
			Inbackground++;
		else
			signal(SIGINT, SIG_DFL);
		signal(SIGHUP, SIG_IGN);
	  ....end comment out..*/

	/* Process the parameters */
	errorstat=0;
	if(argc<2)
		goto usage;
	while (--argc) {
		cp = *++argv;
		if(*cp == '-') {
			while( *++cp) {
				switch(tolower(*cp)) {
				case 'n':
					Nlmode=TRUE; 
					break;
				case 'o':
					if ((outdrv[1]  = *++cp) == ':') outdrv[1] = *++cp;
					if (outdrv[1] <'0' || outdrv[1] >'8') goto usage;
					outdrv[0] = ':';
					outdrv[2] = '\0';
					break;
				case 'v':
					if (isdigit(*(cp+1))) { 
						dispcnt = atoi(++cp);
						while (*++cp) ; 
						cp--;
					} 
					else dispcnt = 65535;
					break;
				default:
					goto usage;
				} /* switch */
			} /* while */
		}  /* if '-' */
		else if( !npats && argc>0) {
			if(argv[0][0]) {
				patts=wcx(&argc,argv);
				npats=argc;
			}
		}
	}
	if(npats < 1) {
usage:
		fprintf(stderr,"Usage: usq [-nov] file ...\n");
		fprintf(stderr,"        -n Nlmode: remove linefeeds\n");
		fprintf(stderr,"        -o output drive follows\n");
		fprintf (stderr,"        -v view (on stdout) optional # lines follows\n");
		fprintf(stderr,"      *,? allowed in filename(s) (command line only)\n");
		fprintf(stderr,"      example: usq  -o:1 */?Q?:0");

		do {
			fprintf(stderr,"\n*");
			for (i = 0 ; i<32 ; ++i) {
				if ((c = getchar()) == EOF) c = '\n';
				if ((inparg[i] = c) == '\n')  {
					inparg[i] = '\0';
					break;
				}
			}
			if (inparg[0] != '\0') errorstat |= usqueeze(inparg);
		} while (inparg[0] != '\0');
		exit(errorstat != 0);
	}
	for(n=0; n<npats; ++n)
		errorstat |= usqueeze(patts[n]);
	exit(errorstat != 0);
}

/*
	The following code is primarily from typesq.c and utr.c.  Typesq
is a modification of USQ by Dick Greenlaw.  Those modifications (usq
to typesq) were made by Bob Mathias
*/

FILE *in, *out;
usqueeze(fname)
char *fname;
{
	register INT i, c,cc;
	register char *p;
	register INT numnodes;			/* size of decoding tree */
	register UNSIGNED crc;
	UNSIGNED filecrc;
	char origname[PATHLEN];		/* Original file name without drive */
	int status;
	status = 0;
	init_cr(); 
	init_huff(); 
	crc=0;

	caps(fname); /* for TRSDOS..*/
#ifdef TRS4
	setro(); /* avoid possible error from lrl <>256 */
#endif
	if((in=fopen( fname, "r"))==NULL) {
		fprintf(stderr, "usq: can't open %s\n", fname);
		return ERROR;
	}
	linecnt = 0;
	if(portgetw(in) != (INT) RECOGNIZE) {/* Process header */
		fprintf(stderr, "usq: %s is not a SQueezed file\n", fname);
		return(ERROR);
	}
	filecrc = (UNSIGNED) portgetw(in);	/* checksum */
	p = origname;				/* Get original file name */
	do {					/* send it to array */
		*p = getc(in);
	} while(*p++ != '\0');

	numnodes = portgetw(in);
	if(numnodes < 0 || numnodes >= NUMVALS) {
		fprintf(stderr, "usq: %s has invalid decode tree\n", fname);
		fclose(in);
		return(ERROR);
	}
	/* Initialize for possible empty tree (SPEOF only) */
	Dnode[0]._children[0] = -(SPEOF + 1);
	Dnode[0]._children[1] = -(SPEOF + 1);

	for(i = 0; i < numnodes; ++i) {	/* Get decoding tree from file */
		Dnode[i]._children[0] = portgetw(in);
		Dnode[i]._children[1] = portgetw(in);
	}
	/* Get translated output bytes and write file */
	/* for TRSDOS always make all caps..
	 *	if(MakeLCPathname && !IsAnyLower(origname))
	 *		uncaps(origname);
	 *	for(p=origname; *p; ++p)		/  change / to _  /
	 *		if( *p == '/')
	 * 		*p = '_';
	 */
	dos(origname);
	if (!Inbackground)
		fprintf(stderr, "usq: %s -> %s\n",fname,origname);
	if (dispcnt){  /* using stdout  */
		putchar ('\n');
		while (((c = getcr()) != EOF) && (linecnt <dispcnt )) {
			cc = 0x7f & c; /* strip high bit */
			if ((cc < ' ') || (cc > '~'))
				switch (cc) {
				case LF:
					goto next;
				case CR:
					++linecnt;
				case '\f':
					cc = CR;
				case '\t':
					break; /* shld expand */
				default:
					cc = '.';
				}
			putchar (cc);
next: 
			;
		}
		putchar (CR);
	}
	else { /* using file output */
		if((out=fopen(origname, "w"))==NULL) {
			fprintf(stderr, "usq: can't create %s\n", origname);
			return ERROR;
		}
		while ((c = getcr()) != EOF) {
			crc += (UNSIGNED) c;
			if ( c == LF  && Nlmode)
				continue;
			if ((putc(c, out)) == ERROR && ferror (out)) {
				fprintf(stderr,"usq: file write error in %s\n",origname);
				exit (-1);
			}
		}
		status |= 	fclose(in);
		status |= 	fflush(out);
		status |= 	fclose(out);
		if( crc != filecrc ) {
			fprintf(stderr, "usq: bad checksum in %s\n", fname);
			fflush(stdout);
			status = -1;
		}
		return(status);
	}
}
/*** from utr.c - */
/* initialize decoding functions */

init_cr()
{
	Repct = 0;
}

init_huff()
{
	Bpos = 99;	/* force initial read */
}

/* Get bytes with decoding - this decodes repetition,
 * calls getuhuff to decode file stream into byte
 * level code with only repetition encoding.
 *
 * The code is simple passing through of bytes except
 * that DLE is encoded as DLE-zero and other values
 * repeated more than twice are encoded as value-DLE-count.
 */

INT
getcr()
{
	register INT c;

	if(Repct > 0) {
		/* Expanding a repeated char */
		--Repct;
		return(Value);
	} 
	else {
		/* Nothing unusual */
		if((c = getuhuff()) != DLE) {
			/* It's not the special delimiter */
			Value = c;
			if(Value == EOF)
				Repct = LARGE;
			return(Value);
		} 
		else {
			/* Special token */
			if((Repct = getuhuff()) == 0)
				/* DLE, zero represents DLE */
				return(DLE);
			else {
				/* Begin expanding repetition */
				Repct -= 2;	/* 2nd time */
				return(Value);
			}
		}
	}
}
/* Decode file stream into a byte level code with only
 * repetition encoding remaining.
 */

INT
getuhuff()
{
	register INT i;

	/* Follow bit stream in tree to a leaf*/
	i = 0;	/* Start at root of tree */
	do {
		if(++Bpos > 7) {
			if((Curin = getc(in)) == ERROR)
				return(ERROR);
			Bpos = 0;
			/* move a level deeper in tree */
			i = Dnode[i]._children[1 & Curin];
		} 
		else
			i = Dnode[i]._children[1 & (Curin >>= 1)];
	} while(i >= 0);

	/* Decode fake node index to original data value */
	i = -(i + 1);
	/* Decode special endfile token to normal EOF */
	i = (i == SPEOF) ? EOF : i;
	return(i);
}
/*
 * Machine independent getw which always gets bytes in the same order
 *  as the CP/M version of SQ wrote them
 */
INT
portgetw(f)
FILE *f;
{
	register INT c;

	c = getc(f) & 0377;
	return(c | (getc(f) << 8));
}


/* make string s lower case */
uncaps(s)
char *s;
{
	for( ; *s; ++s)
		if(isupper(*s))
			*s = tolower(*s);
}


/*
 * IsAnyLower returns TRUE if string s has lower case letters.
 * IsAnyLower(s)
 * char *s;
 * {
 *	for( ; *s; ++s)
 *		if (islower(*s))
 *			return(TRUE);
 *	return(FALSE);
 *}
 */

/* make string upper case */
caps(s)
char *s;
{

	for( ; *s; ++s)
		if(islower(*s))
			*s = toupper(*s);
}
/* dos - makes string a legal LDOS/TRSDOS filename
 * make all U/C, start w/alpha, ext must start w/alpha
 * 8 char name 3 char ext max, separated by '/'
 * other chars must be alphanum.
*/
dos(s)
char *s;
{
	int k,x;
	caps(s);   /* first make all UC */
	k =x= 0;
	while  (*s)  { /* process to end of string */
		if (k++ == 0  && !isalpha(*s)) *s = 'X'; /* 1st alpha */
		if (*s == '.' || k == 9 ) { /* ext separated by slash */
			*s++ = '/';
			k = 0;
			x = TRUE;       /* flag in ext */
			continue;       /* no other check after / */
		}
		if (!isalnum (*s) || *s == '_' ) *s = 'X'; /* all other alnum */
		++s;
		if (x !=0 && k == 3) *s = '\0'; /* max length */
	}
	if (*outdrv) strcat (--s,outdrv); /* add output drive if given */
}
/* wild card filename entry expansion function
 * call with: argv = wcx(&argc,argv)
 * returns new argv array and modifies argc if
 * wildcards are matched
 * argv[0] is not handled as a special case so
 * that the routine may be called to expand a
 * single filename
 * argv is an array of pointers to strings that
 * may be wildcard filenames
 * argc is the int count of the strings
*/
/*
 * wild card is of the form * / *[.pswd]:* (possibly allow ranges later)
 * ? matches any single char / * matches from posn to end of field
 * if ? or * is in file spec and no drive spec given
 * :* assumed (all drives) 
 * trsdos filenames = 8 char name,3 char ext
 * :d = drive spec
 * could have 240 * 8 entries if * / *:*
 * if no match is found the pattern is left unchanged in the argv array
 */
#define ARGNUM 1000  /* max argc allowed */

wcx (cptr,argv) /* expand wild card filespec */
int *cptr;
char **argv;
{
	FILE *dir;           /* read directory as a file */
	char *targv[ARGNUM]; /* local argv array */
	char buff[20];       /* space to construct filename */
	char drive[3];       /* space for drive spec */
	char dirbuff[32];    /* space for raw directory read */
	char *cp,*qp,c,d,dpat; 
	int i,k,j,n,argn;
	drive[0] = ':';
	drive[2] = '\0';      /* set up as string */
	argn = 0; /* index for new argv array */
	for (k=0 ; k < *cptr; k++)  { /* for each argument passed */
		cp = targv[argn] = argv[k]; /* copy pointer to argument */
		/* look for wc */
		if (index(cp,'*') || index (cp,'?') || *cp == '!')  {  /* has wc */
			/*  have possible filespec, now  cycle through drives */
			i=0;        /* flag no match yet */
			dpat = '?';
			if ((qp=index(cp,':')) && isdigit(*++qp)) dpat = *qp;
			for (d='0' ; d<'8' ; d++) {
				drive[1] = d;
				if (dpat == '?' || d == dpat) {
					strcpy (buff,"DIR/SYS");
					strcat (buff,drive);
					if ((dir = fopen (buff,"r")) == NULL)
						continue; /* skip this one */
					fseek (dir,512L,0);/*skip GAT/HIT */
					while (fread(dirbuff,32,1,dir) > 0) {
						/* directory entries 1st byte means:
						 * bit 3, 1 = inv
						 * bit 4, 1 = alive
						 * bit 6, 1 = system
						 * bit 7, 1 = fxde
						 * so mask all others, look for only bit 4 = 1
						 * to find live, vis, fpde only
						*/
						if((*dirbuff == 0x10 ) && /* live,visfpde */
						( n= wlddir(cp,dirbuff+5,buff))) {
							/* found a match - save copy of filename */
							++i;  /* flag match */
							if ((targv[argn]=malloc(n+3)) == NULL) {
nospace:
								fputs ("wcx: no memory\n",stderr);
								exit (-1);
							}
							strcpy (targv[argn],buff);
							strcat(targv[argn++],drive);
						} /* end if live etc. */
					} /* end while (fread) */
					fclose (dir);
				} /* end if this drive */
			} /* end for (drives) */
			/*          if (i==0) argn++;    keep old pattern */
		} /* end if wildc */
		else argn++; /* if not wildc then count copied oldarg */
	}/*end for (args) */
	/* if wildcards expanded to more than one, must alloc new argv array */
	if (argn > *cptr ) {
		if((argv = malloc (argn * 2)) == NULL)/* should be sizeof (char *) */
			goto nospace;
	}
	for (j=0;j<argn;j++) argv[j] = targv[j];
	*cptr = argn; /* set new argc */
	return (argv);
} /* end function */

/* put matching directory entry in buffer as actual filename */
/* return # of chars used in storeit - 0 means no match */
wlddir (spec,actual,storeit)
char *spec,*actual,*storeit;
{
	char c,d;
	int i,n,notflag,nomatch;
	i=notflag =nomatch= 0;
	if (*spec == '!') {
		notflag=1;
		spec++;
	}
	for (n=0; n<11; n++) { /* max 11 chars moved */
		if (n==8) { /* moving to ext field */
			if (actual [n] != ' ') storeit [i++] = '/';
			if (c== '*') spec++;
			c = *spec;
			if (c == '/' || c== '.') spec++;
		} /* end if  n==8 */
		if ((d = actual[n]) != ' ') storeit[i++] = d;
		c = toupper (*spec);
		switch (c) {
		case '.':  /* allow *.* like other systems */
		case '/':
		case '\0':
		case ':':  /* do not advance on these terminators */
			c = ' ';
			break;
		case '*':
			break;
		default:
			spec++;
			break;
		} /* end switch */
		/* set flags according to specified wc for posn */
		if (!(c == '*' ||   /* matchall */
		c == d ||     /* actual match */
		(c =='?' && d != ' '))) nomatch=1; /* match non-space */
	} /* end for chars */
	if (notflag) nomatch = !nomatch; /* reverse sense of match */
	if (nomatch) return (0); /* skip this one */
	if ( (c = *spec) && c != '.')  c = *++spec;      /* if not end of spec */
	if (c =='.') {           /* password may be here */
		for (n=0;n<6;n++) { 
			if ( (c = *spec++) && c != ':' )
				storeit[i++]=c;
			else break;
		} /* end for */
	} /* end if */
	storeit[i++] = '\0';
	return (i);
} /* end function */

#ifdef TRS4
#asm
; fix problem of OPEN error with lrl >0
setro_:
  DB 0FDH,0E5H          ; PUSH IY
  DB 3EH,65H            ; LD      A,101
  DB 0EFH               ; RST     40
  DB 0FDH,0CBH,12H,0C6H ; SET 0,(IY+'S'-'A')
  DB 0FDH,0E1H          ; POP IY
  DB 0C9H               ; RET
;
#endasm
#endif
